"""
Low-level operations for encoding and decoding claims and capabilities.
"""

from petlib.ec import EcGroup, EcPt
from petlib.bn import Bn
from petlib.pack import encode, decode
from profiled import profiled

from .crypto import compute_vrf, verify_vrf, VrfContainer
from .crypto import PublicParams, LocalParams
from .utils import ensure_binary


def _compute_claim_key(vrf_value, mode='enc'):
    if mode not in ['enc', 'lookup']:
        raise ValueError('Invalid mode')
    pp = PublicParams.get_default()
    size = pp.enc_key_size if mode == 'enc' else pp.lookup_key_size
    mode = ensure_binary(mode)
    return pp.hash_func(b"clm_%s|%s" % (mode, vrf_value)).digest()[:size]


def _compute_capability_key(nonce, shared_secret, claim_label, mode='enc'):
    if mode not in ['enc', 'lookup']:
        ValueError('Invalid mode')
    pp = PublicParams.get_default()
    size = pp.enc_key_size if mode == 'enc' else pp.lookup_key_size
    shared_secret_hash = pp.hash_func(shared_secret.export()).digest()
    nonce = ensure_binary(nonce)
    claim_label = ensure_binary(claim_label)
    mode = ensure_binary(mode)
    return pp.hash_func(b"cap_%s|%s|%s|%s" %
            (mode, nonce, shared_secret_hash, claim_label)) \
            .digest()[:size]


def _salt_label(nonce, claim_label):
    nonce = ensure_binary(nonce)
    claim_label = ensure_binary(claim_label)
    return b"lab_%s.%s" % (nonce, claim_label)


# TODO: Remove. Temporary fix for gdanezis/petlib#16
from petlib.bindings import _FFI
def _fix_bytes(tag):
    return bytes(_FFI.buffer(tag)[:])


@profiled
def get_capability_lookup_key(owner_dh_pk, nonce, claim_label):
    """Compute capability lookup key.

    :param petlib.EcPt owner_dh_pk: Owner's DH public key
    :param bytes nonce: Nonce
    :param bytes claim_label: Corresponding claim label
    """
    nonce = ensure_binary(nonce)
    claim_label = ensure_binary(claim_label)
    pp = PublicParams.get_default()
    params = LocalParams.get_default()
    shared_secret = params.dh.sk * owner_dh_pk
    shared_secret_key = pp.hash_func(shared_secret.export()).digest()
    return _compute_capability_key(
            nonce, shared_secret, claim_label, mode='lookup')


@profiled
def encode_claim(nonce, claim_label, claim_content):
    """Encode claim.

    :param bytes nonce: Nonce
    :param bytes claim_label: Claim label
    :param bytes claim_content: Claim content
    """
    nonce = ensure_binary(nonce)
    claim_label = ensure_binary(claim_label)
    claim_content = ensure_binary(claim_content)

    pp = PublicParams.get_default()
    salted_label = _salt_label(nonce, claim_label)
    vrf = compute_vrf(salted_label)
    lookup_key = _compute_claim_key(vrf.value, mode='lookup')
    enc_key = _compute_claim_key(vrf.value, mode='enc')

    claim = encode([vrf.proof, claim_content])
    enc_body, tag = pp.enc_cipher.quick_gcm_enc(
            enc_key, b"\x00"*pp.enc_key_size, claim)
    tag = _fix_bytes(tag)

    enc_claim = encode([enc_body, tag])
    return (vrf.value, lookup_key, enc_claim)


@profiled
def decode_claim(owner_vrf_pk, nonce, claim_label, vrf_value, encrypted_claim):
    """Decode claim.

    :param petlib.EcPt owner_vrf_pk: Owner's VRF public key
    :param bytes nonce: Nonce
    :param bytes claim_label: Claim label
    :param bytes vrf_value: Exported VRF value (hash)
    :param bytes encrypted_claim: Claim content
    """
    claim_label = ensure_binary(claim_label)

    pp = PublicParams.get_default()
    cipher = pp.enc_cipher
    salted_label = _salt_label(nonce, claim_label)
    (encrypted_body, tag) = decode(encrypted_claim)

    lookup_key = _compute_claim_key(vrf_value, mode='lookup')
    enc_key = _compute_claim_key(vrf_value, mode='enc')
    raw_body = cipher.quick_gcm_dec(
            enc_key, b"\x00"*pp.enc_key_size, encrypted_body, tag)
    (proof, claim_content) = decode(raw_body)

    vrf = VrfContainer(value=vrf_value, proof=proof)
    if not verify_vrf(owner_vrf_pk, vrf, salted_label):
        raise Exception("Wrong VRF value")

    return claim_content


@profiled
def encode_capability(reader_dh_pk, nonce, claim_label, vrf_value):
    """Encode capability.

    :param petlib.EcPt reader_dh_pk: Reader's VRF public key
    :param bytes nonce: Nonce
    :param bytes claim_label: Corresponding claim label
    :param bytes vrf_value: Exported VRF value (hash)
    """
    claim_label = ensure_binary(claim_label)
    pp = PublicParams.get_default()
    cipher = pp.enc_cipher
    params = LocalParams.get_default()
    shared_secret = params.dh.sk * reader_dh_pk

    lookup_key = _compute_capability_key(
            nonce, shared_secret, claim_label, mode='lookup')
    enc_key = _compute_capability_key(
            nonce, shared_secret, claim_label, mode='enc')

    enc_body, tag = cipher.quick_gcm_enc(
            enc_key, b"\x00"*pp.enc_key_size, vrf_value)
    tag = _fix_bytes(tag)

    return lookup_key, encode([enc_body, tag])


@profiled
def decode_capability(owner_dh_pk, nonce, claim_label, encrypted_capability):
    """Decode capability.

    :param petlib.EcPt owner_dh_pk: Owder's VRF public key
    :param bytes nonce: Nonce
    :param bytes claim_label: Corresponding claim label
    :param bytes encrypted_capability: Encrypted capability
    """
    pp = PublicParams.get_default()
    cipher = pp.enc_cipher
    params = LocalParams.get_default()
    shared_secret = params.dh.sk * owner_dh_pk
    enc_key = _compute_capability_key(
            nonce, shared_secret, claim_label, mode='enc')
    enc_body, tag = decode(encrypted_capability)
    vrf_value = cipher.quick_gcm_dec(
            enc_key, b"\x00"*pp.enc_key_size, enc_body, tag)
    claim_lookup_key = _compute_claim_key(vrf_value, mode='lookup')
    return vrf_value, claim_lookup_key




"""
Containers for key material.
"""

import base64

from hashlib import sha256

from attr import asdict, attrs, attrib, Factory
from defaultcontext import with_default_context
from petlib.cipher import Cipher
from petlib.ec import EcGroup, EcPt
from petlib.pack import encode, decode

from claimchain.utils import pet2ascii, ascii2pet


@with_default_context(use_empty_init=True)
@attrs
class PublicParams(object):
    """Public parameters of the system."""
    ec_group = attrib(default=Factory(EcGroup))
    hash_func = attrib(default=Factory(lambda: sha256))
    enc_cipher = attrib(default=Factory(lambda: Cipher("aes-128-gcm")))
    enc_key_size = attrib(default=16)
    lookup_key_size = attrib(default=8)
    nonce_size = attrib(default=16)


@attrs
class Keypair(object):
    """Asymmetric key pair.

    :param pk: Public key
    :param sk: Private key
    """
    pk = attrib()
    sk = attrib(default=None)

    @staticmethod
    def generate():
        """Generate a key pair."""
        pp = PublicParams.get_default()
        G = pp.ec_group
        s = G.order().random()
        return Keypair(sk=s, pk=s * G.generator())


@with_default_context
@attrs
class LocalParams(object):
    """ClaimChain user's cryptographic material.

    :param Keypair vrf: VRF key pair
    :param Keypair sig: Signing key pair
    :param Keypair dh: DH key pair
    :param Keypair rescue: Rescue key pair (not used)
    """
    vrf = attrib(default=None)
    sig = attrib(default=None)
    dh = attrib(default=None)
    rescue = attrib(default=None)

    @staticmethod
    def generate():
        """Generate key pairs."""
        return LocalParams(
            vrf = Keypair.generate(),
            sig = Keypair.generate(),
            dh = Keypair.generate(),
            rescue = Keypair.generate()
        )

    def public_export(self):
        """Export public keys to dictionary."""
        return self._export(private=False)

    def private_export(self):
        """Export public and private keys to dictionary."""
        return self._export(private=True)

    def _export(self, private=False):
        result = {}
        for name, attr in asdict(self, recurse=False).items():
            if isinstance(attr, Keypair):
                result[name + '_pk'] = pet2ascii(attr.pk)
                if private:
                    result[name + '_sk'] = pet2ascii(attr.sk)
        return result

    @staticmethod
    def from_dict(exported):
        """Import from dictionary.

        :param dict exported: Exported params
        """
        def maybe_decode(encoded_point):
            if encoded_point is not None:
                return ascii2pet(encoded_point)

        def maybe_load_keypair(prefix):
            keypair = Keypair(
                    pk = maybe_decode(exported.get(prefix + '_pk')),
                    sk = maybe_decode(exported.get(prefix + '_sk')))
            if keypair.pk is not None or keypair.sk is not None:
                return keypair

        params = LocalParams()
        params.vrf = maybe_load_keypair('vrf')
        params.sig = maybe_load_keypair('sig')
        params.dh = maybe_load_keypair('dh')
        params.rescue = maybe_load_keypair('rescue')
        return params




import six
from base58 import b58encode, b58decode
from petlib.pack import encode, decode


def ensure_binary(s):
    """
    >>> ensure_binary(b"test")
    b'test'
    >>> ensure_binary(u"test")
    b'test'
    """
    if not isinstance(s, six.binary_type):
        s = s.encode('utf-8')
    return s


def ensure_text(s):
    if not isinstance(s, six.text_type):
        s = s.decode("ascii")
    return s


def bytes2ascii(s):
    """
    >>> bytes2ascii(b"test")
    '3yZe7d'
    """
    return b58encode(s)


def ascii2bytes(s):
    """
    >>> ascii2bytes('3yZe7d')
    b'test'
    """
    return b58decode(ensure_binary(s))


def pet2ascii(p):
    """
    >>> from petlib.ec import EcGroup, EcPt
    >>> G = EcGroup()
    >>> pt = EcPt(G)
    >>> pet2ascii(pt)
    '3Xw3vNAdCmDLs'
    """
    return ensure_text(b58encode(encode(p)))


def ascii2pet(s):
    """
    >>> ascii2pet('3Xw3vNAdCmDLs')
    EcPt(00)
    """
    return decode(b58decode(ensure_binary(s)))



import json
import pytest

from petlib.ec import EcPt

from claimchain.crypto.params import PublicParams, LocalParams


@pytest.fixture
def local_params():
    with LocalParams.generate().as_default() as params:
        yield params


def test_local_params_public_export(local_params):
    G = PublicParams.get_default().ec_group
    exported = local_params.public_export()

    for key in exported:
        assert not key.endswith("_sk")

    local_params1 = LocalParams.from_dict(exported)

    assert local_params1.vrf.pk == local_params.vrf.pk
    assert local_params1.sig.pk == local_params.sig.pk
    assert local_params1.dh.pk == local_params.dh.pk
    assert local_params1.rescue.pk == local_params.rescue.pk
    assert local_params1.vrf.sk is None
    assert local_params1.sig.sk is None
    assert local_params1.dh.sk is None
    assert local_params1.rescue.sk is None


def test_local_params_private_export(local_params):
    G = PublicParams.get_default().ec_group
    public_exported = local_params.public_export()
    private_exported = local_params.private_export()

    assert public_exported != private_exported
    for key, val in public_exported.items():
        assert private_exported[key] == val

    # check that export contains no bytes and is json dumpable
    data = json.dumps(private_exported)
    imp = json.loads(data)

    local_params1 = LocalParams.from_dict(imp)

    assert local_params1.vrf.sk == local_params.vrf.sk
    assert local_params1.sig.sk == local_params.sig.sk



from claimchain.crypto.params import LocalParams
from claimchain import State, View
from hippiehug import Chain
from claimchain.utils import pet2ascii

def test_read_claim_from_other_chain():
    for i in range(1,100):
        alice_params = LocalParams.generate()
        alice_state = State()
        alice_store = {}
        alice_chain = Chain(alice_store, None)
        alice_state.identity_info = "Hi, I'm " + pet2ascii(alice_params.dh.pk)
        with alice_params.as_default():
            alice_head = alice_state.commit(alice_chain)
        alice_chain = Chain(alice_store, alice_head)

        bob_params = LocalParams.generate()
        bob_state = State()
        bob_store = {}
        bob_chain = Chain(bob_store, None)
        bob_state.identity_info = "Hi, I'm " + pet2ascii(bob_params.dh.pk)
        with bob_params.as_default():
            bob_head = bob_state.commit(bob_chain)
        bob_chain = Chain(bob_store, bob_head)

        bob_pk = bob_params.dh.pk

        with alice_params.as_default():
            alice_state[b"bobs_key"] =  b"123abc"
            alice_state.grant_access(bob_pk, [b"bobs_key"])
            alice_head = alice_state.commit(alice_chain)
        alice_chain = Chain(alice_store, alice_head)

        with alice_params.as_default():
            value = View(alice_chain)[b'bobs_key'].decode('utf-8')

        assert value == "123abc"

        with bob_params.as_default():
            value = View(alice_chain)[b'bobs_key'].decode('utf-8')

        assert value == "123abc"



import pytest

import hippiehug
from petlib.pack import encode, decode

from claimchain.state import State, View, Payload
from claimchain.core import get_capability_lookup_key
from claimchain.core import _compute_claim_key, _salt_label
from claimchain.crypto import PublicParams, LocalParams, compute_vrf
from claimchain.utils import ascii2bytes
from claimchain.utils import Tree, ObjectStore


@pytest.fixture(scope="module", autouse=True)
def local_params():
    with LocalParams.generate().as_default() as params:
        yield params


@pytest.fixture
def state():
    return State()


def test_build_payload(local_params):
    payload = Payload.build(Tree(), b'nonce')
    assert payload.mtr_hash is None


def test_parse_payload(local_params):
    payload = Payload.build(Tree(), b'nonce')
    exported = payload.export()
    payload1 = Payload.from_dict(exported)

    # Check if original dicitonary is not parsed
    assert isinstance(exported['metadata'], dict)


def test_add_claim(state):
    state["marios"] = "test"
    assert state["marios"] == "test"


def test_clear(state):
    state["marios"] = "test"
    state.clear()
    with pytest.raises(KeyError):
        state["marios"]


def test_grant_access(state):
    state["marios"] = "test"
    state["carmela"] = "test"

    reader_params = LocalParams.generate()
    reader_pk = reader_params.dh.pk
    state.grant_access(reader_params.dh.pk, ["marios", "carmela"])
    assert set(state.get_capabilities(reader_pk)) == {"marios", "carmela"}

    state.revoke_access(reader_params.dh.pk, ["marios"])
    assert set(state.get_capabilities(reader_pk)) == {"carmela"}


def commit_claims(state, claims, caps=None):
    for label, content in claims:
        state[label] = content
    for reader_dh_pk, label in (caps or []):
        state.grant_access(reader_dh_pk, label)

    store = {}
    chain = hippiehug.Chain(store)
    head = state.commit(chain)

    block = store[head]
    block_content = block.items[0]
    nonce = ascii2bytes(block_content['nonce'])

    # Get associated Merkle tree
    mtr_hash = ascii2bytes(block_content['mtr_hash'])
    tree = hippiehug.Tree(store, root_hash=mtr_hash)
    return nonce, chain, tree


def test_tree_contains_claim_lookup_key(state):
    nonce, chain, tree = commit_claims(state, [("marios", "test")])

    salted_label = _salt_label(nonce, "marios")
    vrf = compute_vrf(salted_label)
    pp = PublicParams.get_default()

    lookup_key = _compute_claim_key(vrf.value, mode='lookup')
    root, evidence = tree.evidence(key=lookup_key)
    leaf = evidence[-1]
    assert leaf.key == lookup_key


def test_tree_contains_cap_lookup_key(state):
    reader_params = LocalParams.generate()
    reader_pk = reader_params.dh.pk
    state.grant_access(reader_pk, ["marios"])

    nonce, chain, tree = commit_claims(state, [("marios", "test")])

    owner_pk = LocalParams.get_default().dh.pk
    with reader_params.as_default():
        lookup_key = get_capability_lookup_key(owner_pk, nonce, "marios")

    root, evidence = tree.evidence(key=lookup_key)
    leaf = evidence[-1]
    assert leaf.key == lookup_key


def test_evidence(state):
    reader_params = LocalParams.generate()
    reader_pk = reader_params.dh.pk
    state.grant_access(reader_pk, ["marios"])

    nonce, chain, tree = commit_claims(state, [("marios", "test")])
    evidence = state.compute_evidence_keys(reader_pk, "marios")
    assert len(evidence) > 0

    evidence_store = ObjectStore({k: tree.store[k] for k in evidence})
    verification_tree = Tree(object_store=evidence_store, root_hash=tree.root())

    with reader_params.as_default():
        view = View(source_chain=chain, source_tree=verification_tree)
        assert view["marios"] == b"test"


def test_view_missing_and_non_existent_label(state):
    reader_params = LocalParams.generate()
    _, chain, tree = commit_claims(state,
            [("marios", "test1"), ("bogdan", "test2")],
            [(reader_params.dh.pk, ["marios"])])

    with reader_params.as_default():
        view = View(chain)
        with pytest.raises(KeyError):
            view["george"]
        with pytest.raises(KeyError):
            view["bogdan"]


def test_view_label_retrieval(state):
    reader_params = LocalParams.generate()
    _, chain, tree = commit_claims(state,
            [("marios", "test1"), ("bogdan", "test2")],
            [(reader_params.dh.pk, ["marios", "bogdan"])])

    own_view = View(chain)
    assert own_view["marios"] == b"test1"
    assert own_view["bogdan"] == b"test2"

    with reader_params.as_default():
        view = View(chain)
        assert view["marios"] == b"test1"
        assert view["bogdan"] == b"test2"



from petlib.ecdsa import do_ecdsa_setup, do_ecdsa_sign, do_ecdsa_verify

from . import PublicParams, LocalParams


def sign(message):
    """Sign a message.

    :param bytes message: Message
    :return: Tuple of bignums (``petlib.bn.Bn``)
    """
    pp = PublicParams.get_default()
    params = LocalParams.get_default()
    G = pp.ec_group
    digest = pp.hash_func(message).digest()
    kinv_rp = do_ecdsa_setup(G, params.sig.sk)
    sig = do_ecdsa_sign(G, params.sig.sk, digest, kinv_rp=kinv_rp)
    return sig


def verify_signature(sig_pk, sig, message):
    """Verify a signature.

    :param petlib.EcPt sig_pk: Signature verification key
    :param sig: Signature
    :type sig: tuple of bignums (``petlib.bn.Bn``)
    :param bytes message: Message
    """
    pp = PublicParams.get_default()
    G = pp.ec_group
    digest = pp.hash_func(message).digest()
    return do_ecdsa_verify(G, sig_pk, sig, digest)



# -*- coding: utf-8
"""
High-level ClaimChain interface.
"""

import os
import warnings
from time import time
from base64 import b64encode
from hashlib import sha256
from collections import defaultdict

from attr import attrs, attrib, asdict, Factory
from profiled import profiled

from hippiehug import Chain
from hippiehug import Tree

from .core import get_capability_lookup_key
from .core import encode_capability, decode_capability
from .core import encode_claim, decode_claim
from .core import _compute_claim_key
from .crypto import PublicParams, LocalParams
from .crypto import sign, verify_signature
from .utils import bytes2ascii, ascii2bytes, pet2ascii, ascii2pet
from .utils import cached_property
from .utils import Tree, Blob, ObjectStore


PROTOCOL_VERSION = 1


@attrs
class Metadata(object):
    """Block metadata.

    :param params: Owner's cryptographic parameters.
    :param identity_info: Owner's identity info (public key)
    """
    params = attrib()
    identity_info = attrib(default=None)


@attrs
class Payload(object):
    """Block payload.

    :param bytes mtr_hash: Hash of the Merkle tree root
    :param Metadata metadata: Block's metadata
    :param bytes nonce: Nonce
    :param timestamp: Unix-format timestamp
    :param int version: Protocol version
    """

    mtr_hash  = attrib()
    metadata  = attrib()
    nonce     = attrib(default=False)
    timestamp = attrib(default=Factory(lambda: time()))
    version   = attrib(default=PROTOCOL_VERSION)

    @staticmethod
    def build(tree, nonce, identity_info=None):
        """Build a payload.

        :param tree: Tree object
        :param bytes nonce: Nonce
        :param identity_info: Owner's identity info (public key)
        """
        metadata = Metadata(
                params=LocalParams.get_default().public_export(),
                identity_info=identity_info)
        if tree.root_hash is not None:
            mtr_hash = bytes2ascii(tree.root_hash)
        else:
            mtr_hash = None
        return Payload(metadata=metadata,
                       mtr_hash=mtr_hash,
                       nonce=bytes2ascii(nonce))

    @staticmethod
    def from_dict(exported):
        """Import payload from dictionary.

        :param dict exported: Exported payload.
        """
        raw_metadata = exported["metadata"]
        raw_payload = dict(exported)
        raw_payload['metadata'] = Metadata(**raw_metadata)
        return Payload(**raw_payload)

    def export(self):
        """Export to dictionary."""
        return asdict(self)


@profiled
def _build_tree(store, enc_items_map):
    if not isinstance(store, ObjectStore):
        store = ObjectStore(store)
    tree = Tree(store)
    enc_blob_map = {key: Blob(enc_item)
                    for key, enc_item in enc_items_map.items()
                    if not isinstance(enc_item, Blob)}
    tree.update(enc_blob_map)
    return tree


def _sign_block(block):
    sig = sign(block.hash())
    block.aux = pet2ascii(sig)


class State(object):
    """ClaimChain owner state.

    :param identity_info: Owner's identity info (public key)
    """

    def __init__(self, identity_info=None):
        self.identity_info = identity_info

        self._claim_content_by_label = {}
        self._caps_by_reader_pk = defaultdict(set)
        self._enc_items_map = {}
        self._vrf_value_by_label = {}
        self._payload = None
        self._tree = None

    @property
    def tree(self):
        """Corresponding Merkle tree holding the claims and capabilities."""
        if self._tree is None:
            raise ValueError('State not committed yet.')
        return self._tree

    def commit(self, target_chain, tree_store=None, nonce=None):
        """Commit state to a chain.

        Constructs a new block and appends to a chain.

        :param hippiehug.Chain target_chain: Chain to which a block will be
                appended.
        :param utils.ObjectStore tree_store: Object store to hold tree nodes.
        :param bytes nonce: Nonce to include in the new block.
        """
        if tree_store is None:
            tree_store = target_chain.store
        self._nonce = nonce = \
                nonce or os.urandom(PublicParams.get_default().nonce_size)

        # Encode claims
        enc_items_map = {}
        vrf_value_by_label = {}
        for claim_label, claim_content in self._claim_content_by_label.items():
            vrf_value, lookup_key, enc_claim = encode_claim(
                    nonce, claim_label, claim_content)
            enc_items_map[lookup_key] = enc_claim
            vrf_value_by_label[claim_label] = vrf_value

        # Encode capabilities
        for reader_dh_pk, caps in self._caps_by_reader_pk.items():
            for claim_label in caps:
                try:
                    vrf_value = vrf_value_by_label[claim_label]
                except KeyError:
                    warnings.warn("VRF for %s not computed. "
                                  "Skipping adding a capability." \
                                  % claim_label)
                    break
                lookup_key, enc_cap = encode_capability(
                        reader_dh_pk, nonce, claim_label, vrf_value)
                enc_items_map[lookup_key] = enc_cap

        # Put all the encrypted items in a new tree
        tree = _build_tree(tree_store, enc_items_map)

        # Construct payload
        payload = Payload.build(
                tree=tree,
                identity_info=self.identity_info,
                nonce=nonce)
        target_chain.multi_add([payload.export()], pre_commit_fn=_sign_block)

        self._payload = payload
        self._tree = tree
        self._enc_items_map = enc_items_map
        self._vrf_value_by_label = vrf_value_by_label

        return target_chain.head

    def compute_evidence_keys(self, reader_dh_pk, claim_label):
        """List hashes of all nodes that prove inclusion of a claim label.

        :param petlib.EcPt reader_dh_pk: Reader's DH public key
        :param bytes claim_label: Claim label
        """
        try:
            vrf_value = self._vrf_value_by_label[claim_label]
            cap_lookup_key = get_capability_lookup_key(
                    reader_dh_pk, self._nonce, claim_label)

            # Compute capability entry evidence
            _, raw_cap_evidence = self.tree.evidence(cap_lookup_key)
            claim_lookup_key = _compute_claim_key(vrf_value, mode='lookup')

            # Compute claim evidence
            _, raw_claim_evidence = self.tree.evidence(claim_lookup_key)
            object_keys = {obj.hid for obj in raw_cap_evidence} | \
                          {obj.hid for obj in raw_claim_evidence}

            # Add encoded capability and encoded claim value
            encoded_cap_hash = raw_cap_evidence[-1].item
            encoded_claim_hash = raw_claim_evidence[-1].item
            return object_keys | {encoded_claim_hash} | {encoded_cap_hash}
        except KeyError:
            return set()

    def clear(self):
        """Clear buffer."""
        self._claim_content_by_label.clear()
        self._caps_by_reader_pk.clear()

        self._enc_items_map.clear()
        self._vrf_value_by_label.clear()
        self._payload = None
        self._tree = None

    def __getitem__(self, label):
        """Get queued claim by label.

        :param label: Claim label
        """
        return self._claim_content_by_label[label]

    def __setitem__(self, claim_label, claim_content):
        """Add a claim with given label and content to be committed.

        :param bytes claim_label: Claim label
        :param bytes claim_content: Claim content
        """
        self._claim_content_by_label[claim_label] = claim_content

    def grant_access(self, reader_dh_pk, claim_labels):
        """Grant access for given claims a reader.

        :param petlib.EcPt reader_dh_pk: Reader's DH public key
        :param iterable claim_labels: List of claim labels
        """
        self._caps_by_reader_pk[reader_dh_pk].update(set(claim_labels))

    def revoke_access(self, reader_dh_pk, claim_labels):
        """Revoke access for given claims to a reader.

        :param petlib.EcPt reader_dh_pk: Reader's DH public key
        :param iterable claim_labels: List of claim labels
        """
        self._caps_by_reader_pk[reader_dh_pk].difference_update(claim_labels)

    def get_capabilities(self, reader_dh_pk):
        """List all labels accessibly by a reader.

        :param petlib.EcPt reader_dh_pk: Reader's DH public key
        """
        return list(self._caps_by_reader_pk[reader_dh_pk])


class View(object):
    """View of an existing ClaimChain."""

    def __init__(self, source_chain, source_tree=None):
        """
        :param hippiehug.Chain source_chain: Chain to view
        :param utils.Tree source_tree: Tree object if available
        """
        self._viewer_params = LocalParams.get_default()
        self.chain = source_chain
        self._latest_block = self.chain.store[self.chain.head]
        self._nonce = ascii2bytes(self.payload.nonce)
        if self.payload.mtr_hash is not None:
            self.tree = source_tree or Tree(
                    object_store=ObjectStore(self.chain.store),
                    root_hash=ascii2bytes(self.payload.mtr_hash))

            if ascii2bytes(self.payload.mtr_hash) != self.tree.root_hash:
                raise ValueError("Supplied tree doesn't match MTR in the chain.")

    @property
    def head(self):
        """Chain's head (latest block hash)."""
        return self.chain.head

    @cached_property
    def payload(self):
        """Chain's latest block payload."""
        return Payload.from_dict(self._latest_block.items[0])

    @cached_property
    def params(self):
        """Cryptographic params of the chain owner."""
        return LocalParams.from_dict(self.payload.metadata.params)

    # TODO: This validation is incorrect for any block but the genesis
    def validate(self):
        """Validate the chain.

        .. note ::
            Don't use this method. It is broken. ¯\\_(ツ)_/¯
        """
        owner_sig_pk = self.params.sig.pk
        raw_sig_backup = self._latest_block.aux
        sig = ascii2pet(raw_sig_backup)
        self._latest_block.aux = None
        if not verify_signature(owner_sig_pk, sig, self._latest_block.hash()):
            self._latest_block.aux = raw_sig_backup
            raise ValueError("Invalid signature.")
        self._latest_block.aux = raw_sig_backup

    def _lookup_capability(self, claim_label):
        cap_lookup_key = get_capability_lookup_key(
                self.params.dh.pk, self._nonce, claim_label)
        try:
            cap = self.tree[cap_lookup_key]
        except KeyError:
            raise KeyError("Label does not exist or you don't have "
                           "permission to read.")
        except AttributeError:
            raise ValueError("The chain does not have a claim map.")
        return decode_capability(self.params.dh.pk, self._nonce,
                                 claim_label, cap)

    def _lookup_claim(self, claim_label, vrf_value, claim_lookup_key):
        try:
            enc_claim = self.tree[claim_lookup_key]
        except KeyError:
            raise KeyError("Claim not found, but permission to read the label "
                           "exists.")
        except AttributeError:
            raise ValueError("The chain does not have a claim map.")
        return decode_claim(self.params.vrf.pk, self._nonce,
                            claim_label, vrf_value, enc_claim)

    def __getitem__(self, claim_label):
        """Get claim by label.

        :param bytes claim_label: Claim label
        :raises: ``KeyError`` if claim not found or not accessible
        """
        if self._viewer_params.vrf.pk == self.params.vrf.pk:
            vrf_value, claim_lookup_key, enc_claim = encode_claim(
                    self._nonce, claim_label, "")
            claim = self._lookup_claim(claim_label, vrf_value, claim_lookup_key)
        else:
            vrf_value, claim_lookup_key = self._lookup_capability(claim_label)
            claim = self._lookup_claim(claim_label, vrf_value, claim_lookup_key)
        return claim

    def get(self, claim_label):
        """Get claim by label.

        :param bytes claim_label: Claim label
        :return: Claim or ``None`` if not found or not accessible.
        """
        try:
            return self[claim_label]
        except KeyError:
            return None
        except ValueError:
            return None

    def __hash__(self):
        return hash(self.head)



import hippiehug
from hippiehug.Utils import binary_hash

from .encodings import ensure_binary


# TODO: Move to hippiehug 1.0
# TODO: Move the tests out of doctests


class Blob(bytes):
    @property
    def hid(self):
        return binary_hash(self)


def _check_hash(key, value):
    if value.hid != key:
        raise ValueError('Hash of the value is not the lookup key')


class ObjectStore(object):
    """
    >>> store = ObjectStore()
    >>> blob = Blob(b'test')
    >>> store.add(blob)
    >>> store[blob.hid]
    b'test'
    """
    def __init__(self, backend=None):
        self._backend = backend
        if backend is None:
            self._backend = {}

        # Check hashes if it is a plain dictionary
        if not isinstance(self._backend, ObjectStore):
            for lookup_key, value in self._backend.items():
                _check_hash(lookup_key, value)
        else:
            # If input is another ObjectStore, unwrap the
            # underlying dictionary
            self._backend = self._backend._backend

    def __getitem__(self, lookup_key):
        value = self._backend[lookup_key]
        _check_hash(lookup_key, value)
        return value

    def get(self, lookup_key):
        try:
            return self._backend[lookup_key]
        except KeyError:
            return None

    def __setitem__(self, lookup_key, value):
        _check_hash(lookup_key, value)
        self._backend[lookup_key] = value

    def keys(self):
        return self._backend.keys()

    def values(self):
        return self._backend.values()

    def items(self):
        return self._backend.items()

    def add(self, value):
        lookup_key = value.hid
        self._backend[lookup_key] = value



# TODO: (Higher priority) move this to hippiehug classes themselves
def serialize_object(obj):
    """
    Serialize blobs, hippiehug tree nodes, and hippiehug chain blocks

    >>> block = hippiehug.Block([])
    >>> len(serialize_object(block))
    4
    >>> node = hippiehug.Nodes.Branch(b'pivot', b'left', b'right')
    >>> len(serialize_object(node))
    3
    >>> leaf = hippiehug.Nodes.Leaf(b'item', b'key')
    >>> len(serialize_object(leaf))
    2
    >>> blob = Blob(b'content')
    >>> serialize_object(blob) == blob
    True

    .. warning::

    There is no guarantee this is in sync with hippiehug, i.e., this
    is the serialization hippiehug employs internally. This will eventually
    move inside the hippiehug library.

    """
    if isinstance(obj, hippiehug.Nodes.Leaf):
        return (obj.key, obj.item)
    elif isinstance(obj, hippiehug.Nodes.Branch):
        return (obj.pivot, obj.left_branch, obj.right_branch)
    elif isinstance(obj, hippiehug.Block):
        return (obj.index, obj.fingers, obj.items, obj.aux)
    elif isinstance(obj, Blob):
        return obj


class Chain(object):
    def __init__(self, object_store=None):
        self.object_store = object_store or ObjectStore()


class Tree(object):
    """
    Wrapper to enable map interface on top of Hippiehug trees

    >>> tree = Tree()
    >>> b'label' in tree
    False
    >>> tree[b'label']
    Traceback (most recent call last):
    KeyError: b'label'
    >>> tree[b'label'] = Blob(b'test')
    >>> tree[b'label']
    b'test'
    >>> b'label' in tree
    True

    Creating a tree from existing storage:

    >>> tree1 = Tree(tree.object_store, root_hash=tree.root_hash)
    >>> b'label' in tree1
    True
    >>> tree.evidence(b'label') is not None
    True

    Adding multiple items at once:

    >>> b'label1' in tree
    False
    >>> b'label2' in tree
    False
    >>> tree.update({'label1': Blob(b'test1'), 'label2': Blob(b'test2')})
    >>> b'label1' in tree
    True
    >>> b'label2' in tree
    True
    >>> tree[b'label1']
    b'test1'
    >>> tree[b'label2']
    b'test2'
    """
    def __init__(self, object_store=None, root_hash=None):
        self.object_store = object_store
        if object_store is None:
            self.object_store = ObjectStore()
        self.tree = hippiehug.Tree(self.object_store, root_hash=root_hash)

    @property
    def root_hash(self):
        return self.tree.root()

    def __getitem__(self, lookup_key):
        lookup_key = ensure_binary(lookup_key)
        _, evidence = self.evidence(lookup_key)
        if not evidence or evidence[-1].key != lookup_key:
            raise KeyError(lookup_key)
        value_hash = evidence[-1].item
        return self.tree.store[value_hash]

    def __setitem__(self, lookup_key, value):
        """
        Add value with given lookup key.

        TODO: Add transactions. If this fails or stops at some point,
        storage will be left in a screwed up state.

        :param value: An object with ``hid`` property (e.g. ``Blob`` object)
        """
        lookup_key = ensure_binary(lookup_key)
        if not hasattr(value, 'hid'):
            raise TypeError('Value is not a valid object.')

        self.tree.add(key=lookup_key, item=value)
        _, evidence = self.evidence(lookup_key)
        assert self.tree.is_in(value, key=lookup_key)
        value_hash = evidence[-1].item

        # This should only happen if hashing here and in hippiehug
        # become inconsistent because of internal changes in hippiehug
        # TODO: Remove when hashing when changes here are moved to
        # hippiehug
        assert value_hash == value.hid

        self.tree.store[value_hash] = value

    def update(self, items):
        """
        Add multiple values.

        TODO: Add transactions. If this fails or stops at some point,
        storage will be left in a screwed up state.

        :param items: dictionary, where the values are objects with
                      ``hid`` property (e.g. ``Blob`` objects)
        """
        items = {ensure_binary(key): value for key, value in items.items()}
        for value in items.values():
            if not hasattr(value, 'hid'):
                raise TypeError('Value is not a valid object.')
            self.tree.store[value.hid] = value

        if len(items) > 0:
            self.tree.multi_add(list(items.values()), list(items.keys()))

    def __contains__(self, lookup_key):
        lookup_key = ensure_binary(lookup_key)
        _, evidence = self.evidence(lookup_key)
        return evidence != [] and evidence[-1].key == lookup_key

    def evidence(self, lookup_key):
        result = self.tree.evidence(key=lookup_key)
        if not result:
            result = None, []
        return result


def check_evidence(root_hash, evidence, lookup_key):
    """
    >>> tree = Tree()
    >>> tree[b'label'] = Blob(b'test')
    >>> root_hash, evidence = tree.evidence(b'label')
    >>> check_evidence(root_hash, evidence, b'label')
    True
    >>> check_evidence(root_hash, evidence, b'label1')
    False
    """
    lookup_key = ensure_binary(lookup_key)
    store = {node.identity(): node for node in evidence}
    tree = Tree(store, root_hash=root_hash)
    return lookup_key in tree



from petlib.ec import EcGroup

from claimchain.core import encode_claim, decode_claim
from claimchain.core import encode_capability, decode_capability, \
        get_capability_lookup_key
from claimchain.core import _compute_claim_key
from claimchain.crypto import PublicParams, LocalParams


def test_encode_claim_correctness():
    nonce = b"42"
    claim_label = b"george@george.com"
    claim_body = b"This is a test claim"

    with LocalParams.generate().as_default() as params:
        enc = encode_claim(nonce, claim_label, claim_body)
        vrf_value, lookup_key, encrypted_body = enc

        claim2 = decode_claim(params.vrf.pk, nonce, claim_label,
                              vrf_value, encrypted_body)
        assert claim2 == claim_body


def test_encode_cap_correctness():
    owner_params = LocalParams.generate()
    reader_params = LocalParams.generate()

    nonce = b"42"
    claim_label = b"marios@marios.com"
    vrf_value = b"1337"

    with owner_params.as_default():
        lookup_key, encrypted_capability = encode_capability(
                reader_params.dh.pk, nonce, claim_label, vrf_value)

    with reader_params.as_default():
        lookup_key2 = get_capability_lookup_key(
                owner_params.dh.pk, nonce, claim_label)

    assert lookup_key == lookup_key2

    with reader_params.as_default():
        vrf_value2, claim_lookup_key2 = decode_capability(
                owner_params.dh.pk, nonce, claim_label, encrypted_capability)
    assert vrf_value == vrf_value2

    claim_lookup_key = _compute_claim_key(vrf_value, mode='lookup')
    assert claim_lookup_key == claim_lookup_key2



"""
Implementation of a CONIKS's verifiable random function scheme.
"""

from attr import attrs, attrib
from petlib.ec import EcGroup, EcPt
from petlib.bn import Bn
from petlib.pack import encode, decode
from profiled import profiled

from hashlib import sha256

from .params import PublicParams, LocalParams


@attrs
class VrfContainer(object):
    """VRF value (hash) and proof.

    :param bytes value: Exported VRF value (hash)
    :param bytes proof: Exported VRF proof
    """
    value = attrib()
    proof = attrib()


@profiled
def compute_vrf(message):
    """Compute VRF.

    Produces a VRF value (hash) and a proof.

    :param bytes message: Message
    :return: :py:class:`VrfContainer`
    """
    pp = PublicParams.get_default()
    local_params = LocalParams.get_default()

    G = pp.ec_group
    g = G.generator()
    k = local_params.vrf.sk
    pub = local_params.vrf.pk
    h = G.hash_to_point(b"1||" + message)
    v = k * h
    r = G.order().random()
    R = r * g
    Hr = r * h
    s = Bn.from_binary(sha256(encode([g, h, pub, v, R, Hr])).digest())
    t = (r - s * k) % G.order()
    return VrfContainer(value=v.export(), proof=encode((s, t)))


@profiled
def verify_vrf(pub, vrf, message):
    """Verify a VRF.

    Checks whether a VRF value and a proof correspond to the message.

    :param petlib.EcPt pub: VRF public key
    :param VrfContainer vrf: VRF value and proof
    :param bytes message: Message
    """

    pp = PublicParams.get_default()

    G = pp.ec_group
    g = G.generator()
    h = G.hash_to_point(b"1||" + message)
    v = EcPt.from_binary(vrf.value, G)
    s, t = decode(vrf.proof)
    R = t*g + s*pub
    Hr = t*h + s*v
    s2 = Bn.from_binary(sha256(encode([g, h, pub, v, R, Hr])).digest())
    return s2 == s



import os
import time
import json
import random

from os import urandom
from binascii import hexlify
from pprint import pprint
from hashlib import sha256

from hippiehug import Chain
from hippiehug import Tree, Leaf, Branch
from petlib.pack import encode
from petlib.ecdsa import do_ecdsa_setup, do_ecdsa_sign, do_ecdsa_verify
from msgpack import packb

from claimchain.state import Payload, State, View
from claimchain.core import encode_claim, decode_claim
from claimchain.core import encode_capability, decode_capability, get_capability_lookup_key
from claimchain.crypto import LocalParams, PublicParams, sign, Keypair
from claimchain.utils import pet2ascii


def rhex(l):
    return hexlify(urandom(l))[:l]


def generate_test_data(nb_friends=200, nb_per_friend=5):
    labels = [b"%s@%s.com" % (rhex(8), rhex(8)) for _ in range(nb_friends)]
    heads = [urandom(20) for _ in range(nb_friends)]
    pp = PublicParams.get_default()
    G = pp.ec_group

    params_per_friend = [LocalParams.generate() for _ in range(nb_friends)]
    pubkeys = [params.dh.pk for params in params_per_friend]
    privkeys = [params.dh.sk for params in params_per_friend]
    all_data = (labels, heads, pubkeys, privkeys)

    friends_graph = {}
    for f in range(nb_friends):
        cap = random.sample(range(nb_friends), nb_per_friend)
        friends_graph[f] = cap

    return friends_graph, all_data


import pytest


@pytest.mark.skip
def test_e2e_timings():
    friends_graph, all_data = generate_test_data()
    (labels, heads, pubkeys, privkeys) = all_data

    with LocalParams.generate().as_default() as params:
        nonce = os.urandom(PublicParams.get_default().nonce_size)

        # Encode claims
        t0 = time.time()
        c0 = 0
        enc_claims = []
        vrfs = []
        for claim_label, claim_body in zip(labels, heads):
            c0 += 1
            c0 += 1
            enc = encode_claim(nonce, claim_label, claim_body)
            vrf_value, lookup_key, encrypted_claim = enc
            enc_claims += [(lookup_key, encrypted_claim)]
            vrfs += [vrf_value]
        t1 = time.time()
        print("\n\t\tTiming per encoded claim: %1.1f ms" % ((t1-t0) / c0 * 1000))

        # Encode capabilities
        t0 = time.time()
        c0 = 0
        capabilities = []
        cap_index = {}
        for friend in friends_graph:
            friend_dh_pk = pubkeys[friend]
            for fof in friends_graph[friend]:
                c0 += 1
                claim_label = labels[fof]
                vrf_value = vrfs[fof]
                cap_lookup_key, encrypted_cap = encode_capability(
                        friend_dh_pk, nonce, claim_label, vrf_value)
                capabilities += [(cap_lookup_key, encrypted_cap)]
                cap_index[(friend, fof)] = (cap_lookup_key, encrypted_cap)
        t1 = time.time()
        print("\t\tTiming per encoded capab: %1.1f ms" % ((t1-t0) / c0 * 1000))

        data = encode([enc_claims, capabilities])
        print("\t\tData length: %1.1f kb" % (len(data) / 1024.0))

        # Build our non-equivocable tree
        t0 = time.time()
        tree = Tree()
        for lookup_key, enc_item in enc_claims + capabilities:
            tree.add(key=lookup_key, item=enc_item)
            _, evidence = tree.evidence(key=lookup_key)
            assert tree.is_in(enc_item, key=lookup_key)
            enc_item_hash = evidence[-1].item
            tree.store[enc_item_hash] = enc_item

        t1 = time.time()
        print("\t\tTiming for building non-equiv. tree: %1.1f ms" % ((t1-t0) * 1000))

        # Build a chain and a block
        t0 = time.time()
        c0 = 200
        for _ in range(c0):
            chain = Chain(tree.store)
            payload = Payload.build(tree, nonce).export()

            def sign_block(block):
                sig = sign(block.hash())
                block.aux = pet2ascii(sig)

            chain.multi_add([payload], pre_commit_fn=sign_block)

            # Pack block
            block = chain.store[chain.head]
            packed_block = packb(
                    ("S", block.index, block.fingers, block.items, block.aux))

        t1 = time.time()

        print("\t\tTiming for building a block: %1.1f ms" % ((t1-t0) / c0 * 1000))
        print("\t\tPacked block size: %d bytes" % (len(packed_block)))

        t0 = time.time()
        c0 = 0

        # Pick a random reader
        for reader in friends_graph:
            reader_params = LocalParams(
                    dh=Keypair(sk=privkeys[reader], pk=pubkeys[reader]))
            for reader_friend in friends_graph[reader]:
                claim_label = labels[reader_friend]
                with reader_params.as_default():
                    view = View(chain)
                    c0 += 1
                    head = view[labels[reader_friend]]
                    assert head == heads[reader_friend]

        t1 = time.time()
        print("\t\tTiming for retrieving a claim by label: %1.1f ms" %
                ((t1-t0) / c0 * 1000))

        # Pick a target proof to produce
        f1 = random.choice(list(friends_graph.keys()))
        f2 = random.choice(list(friends_graph[f1]))

        (cap_lookup_key, encrypted_cap) = cap_index[(f1, f2)]
        (claim_lookup_key, encrypted_claim) = enc_claims[f2]

        root, e1 = tree.evidence(key=cap_lookup_key)
        _, e2 = tree.evidence(key=claim_lookup_key)

        evidence_store = {e.identity(): e for e in e1+e2}
        t2 = Tree(evidence_store, root)

        assert t2.is_in(key=cap_lookup_key, item=encrypted_cap)
        assert t2.is_in(key=claim_lookup_key, item=encrypted_claim)

        # Serialize evidence:
        evidence = []
        for e in e1 + e2:
            if isinstance(e, Leaf):
                evidence += [(e.key, e.item, tree.store[e.item])]
            elif isinstance(e, Branch):
                evidence += [(e.pivot, e.left_branch, e.right_branch)]
            else:
                pass

        import zlib
        bin_evidence = encode(evidence)
        bin_evidence_compressed = zlib.compress(bin_evidence, 9)
        print("\t\tSize for one proof: %s bytes (compressed %s bytes)" %
                (len(bin_evidence), len(bin_evidence_compressed)))

        print("\t\tPayload:")
        pprint(payload)



import pytest

from claimchain.crypto.params import LocalParams
from claimchain.crypto.sign import sign, verify_signature


@pytest.fixture()
def local_params():
    with LocalParams.generate().as_default() as params:
        yield params


def test_sign_correct(local_params):
    sig = sign(b"test@test.com")
    assert verify_signature(local_params.sig.pk, sig, b"test@test.com")


def test_sign_incorrect_message(local_params):
    sig = sign(b"test@test.com")
    assert not verify_signature(local_params.sig.pk, sig, b"other@test.com")


def test_sign_incorrect_pubkey(local_params):
    sig = sign(b"test@test.com")

    # Generate a distinct public key
    other_params = LocalParams.generate()
    while other_params.sig.pk == local_params.sig.pk:
        other_params = LocalParams.generate()

    assert not verify_signature(other_params.sig.pk, sig,
            b"test@test.com")


def test_sign_nondeterministic(local_params):
    sig1 = sign(b"test@test.com")
    sig2 = sign(b"test@test.com")
    assert sig1 != sig2



from os import path
from codecs import open
from setuptools import setup, find_packages


here = path.abspath(path.dirname(__file__))


with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='claimchain',
    version='0.3.1',
    packages=["claimchain", "claimchain.crypto", "claimchain.utils"],
    license='MIT',
    description='Implementation of ClaimChain, a cryptographic data structure',
    long_description=long_description,
    author=('Bogdan Kulynych (EPFL SPRING Lab), '
            'Marios Isaakidis, George Danezis (UCL)'),
    author_email=('bogdan.kulynych@epfl.ch, '
                  'm.isaakidis@cs.ucl.ac.uk, g.danezis@ucl.ac.uk'),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Security :: Cryptography'
    ],
    install_requires=[
        'six',
        'petlib',
        'pyyaml',
        'attrs',
        'base58',
        'statistics',
        'defaultcontext',
        'hippiehug >= 0.1.3',
        'profiled'
    ],

)




import pytest

from claimchain.crypto.params import LocalParams
from claimchain.crypto.vrf import compute_vrf, verify_vrf


@pytest.fixture()
def local_params():
    with LocalParams.generate().as_default() as params:
        yield params


def test_vrf_correct(local_params):
    vrf = compute_vrf(b"test@test.com")
    assert verify_vrf(local_params.vrf.pk, vrf, b"test@test.com")


def test_vrf_incorrect_message(local_params):
    vrf = compute_vrf(b"test@test.com")
    assert not verify_vrf(local_params.vrf.pk, vrf, b"other@test.com")


def test_vrf_incorrect_pubkey(local_params):
    vrf = compute_vrf(b"test@test.com")

    # Generate a distinct public key
    other_params = LocalParams.generate()
    while other_params.vrf.pk == local_params.vrf.pk:
        other_params = LocalParams.generate()

    assert not verify_vrf(other_params.vrf.pk, vrf,
            b"test@test.com")


def test_vrf_deterministic_value(local_params):
    vrf1 = compute_vrf(b"test@test.com")
    vrf2 = compute_vrf(b"test@test.com")
    assert vrf1.value == vrf2.value

